/*
 * PressureSensor.c
 *
 * Created: 2012-11-16 11:28:52 AM
 *  Author: nrqm
 */ 

#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>
//#include "Arduino.h"
//#include "HardwareSerial.h"

/*#define CLK_HIGH()	(PORTH |= _BV(PH6))
#define CLK_LOW()	(PORTH &= ~_BV(PH6))
#define DATA_HIGH()	(PORTH |= _BV(PH4))
#define DATA_LOW()	(PORTH &= ~_BV(PH4))
#define DATA_BIT()	((PINH & _BV(PH5)) != 0)*/
#define SPI_DDR DDRB	// DDR of SPI port
#define SPI_PORT PORTB	// SPI port
#define SPI_MOSI PB2	// MOSI pin (Master out, Slave in)
#define SPI_MISO PB3	// MISO pin (Master in, Slave out)
#define SPI_SCK PB1		// SCK pin (SPI clock)
#define SPI_SS PB0		// SS pin (Slave Select)
// wait for an SPI read/write operation to complete
#define SPI_WAIT()              while ((SPSR & _BV(SPIF)) == 0);

// bit patterns for reading calib words and sensor data, and for sending reset
#define READ_W1 0b0001110101010000
#define READ_W2 0b0001110101100000
#define READ_W3 0b0001110110010000
#define READ_W4 0b0001110110100000
//#define READ_W_BITS 12

#define READ_D1 0b0000111101000000	// pressure
#define READ_D2 0b0000111100100000	// temperature
//#define READ_D_BITS 10

#define RESET   0b1010101010101010
///#define RESET_BITS 21

typedef struct {
	uint16_t SENST1;
	uint16_t OFFT1;
	uint16_t TCS;
	uint16_t TCO;
	uint16_t Tref;
	uint16_t TEMPSENS;
} coeff_t;

void print(const char* fmt, ...)
{
	char buffer[256];
	va_list args;
	uint8_t size;
	uint8_t i;
	va_start(args, fmt);
	size = vsnprintf(buffer, sizeof(buffer)-1, fmt, args);
	va_end(args);
	
	while ((UCSR0A & _BV(UDRE0)) == 0);	// wait for data register empty
	for (i = 0; i < size; i++)
	{
		UDR0 = buffer[i];
		while ((UCSR0A & _BV(UDRE0)) == 0);
		//UCSR0A |= _BV(TXC0);	// clear flag
	}
}

void send_bytes(uint16_t data)
{
	SPCR &= ~_BV(CPOL);
	SPDR = data >> 8;
	SPI_WAIT();
	SPDR = data & 0xFF;
	SPI_WAIT();
}
uint16_t get_bytes()
{
	uint16_t data;
	SPCR |= _BV(CPOL);
	SPDR = 0;
	SPI_WAIT();
	data = SPDR;
	data <<= 8;
	SPDR = 0;
	SPI_WAIT();
	data |= SPDR;
	return data;
}


void get_coeffs(coeff_t* coefficients)
{
	uint16_t w1, w2, w3, w4;
	
	send_bytes(READ_W1);
	w1 = get_bytes();
	send_bytes(READ_W2);
	w2 = get_bytes();
	send_bytes(READ_W3);
	w3 = get_bytes();
	send_bytes(READ_W4);
	w4 = get_bytes();
	
	print("w1 = %u\tw2 = %u\tw3 = %u\tw4 = %u\n\r", w1, w2, w3, w4);
	
	coefficients->SENST1 = w1 >> 3;
	
	coefficients->OFFT1 = (w1 & 0b111) << 10;
	coefficients->OFFT1 |= w2 >> 6;
	
	coefficients->TCS = w3 >> 6;
	
	coefficients->TCO = w4 >> 7;
	
	coefficients->Tref = (w2 & 0b111111) << 6;
	coefficients->Tref |= w3 & 0b111111;
	
	coefficients->TEMPSENS = w4 & 0b1111111;
}

uint16_t read_temp()
{
	send_bytes(READ_D2);
	//send_bytes(0);
	//while (MISO_HIGH());	// MISO goes low on conversion completion
	_delay_ms(33);
	return get_bytes();
}

int16_t get_temp_diff(coeff_t* coeffs)
{
	uint16_t T = read_temp();
	//print("Raw T reading = %u\r\n", T);
	// find reference temperature
	uint16_t UT1 = 8 * coeffs->Tref + 10000;	// maximum value is 42760
	int16_t dT = T - UT1;		// find difference between temperature reading and reference
	// calculate second-order temperature differential
	int32_t dT_sq = dT;
	dT_sq *= dT;	// dT squared
	dT_sq /= 128;
	dT_sq /= 128;	// looking for (dT/128) * (dT/128)
	if (dT >= 0) dT_sq /= 8;
	else dT_sq /= 2;
	// correct dT using second-order differential
	//print("dT = %d\r\n", dT - dT_sq);
	return dT - dT_sq;
	return 0;
}

int16_t get_temperature(int16_t dT, coeff_t* coeffs)
{
	int32_t acc = dT;
	acc *= (coeffs->TEMPSENS + 100);
	acc /= 2048;
	acc += 200;
	return (uint16_t)acc;
}
uint16_t get_pressure(int16_t dT, coeff_t* coeffs)
{
	int16_t off;	// offset at temperature
	int32_t sens;	// sensitivity at temperature
	uint16_t P;		// pressure value
	int32_t pressure;	// actual pressure in mbar
	
	off = coeffs->TCO - 250;
	off *= dT;		// this shouldn't overflow, but watch out.
	off /= 4096;
	off += coeffs->OFFT1 + 10000;
	//print("off=%d\n\r", off);
	
	sens = coeffs->TCS + 200;
	sens *= dT;
	sens /= 8192;
	sens += coeffs->SENST1 / 2 + 3000;
	send_bytes(READ_D1);
	//send_bytes(0);
	_delay_ms(33);
	P =  get_bytes();
	pressure = P;
	pressure -= off;
	pressure *= sens;
	pressure /= 2048;
	pressure += 1000;
	return (uint16_t)pressure;
}

int main(void)
{
	uint16_t dT;
	uint16_t v;
	coeff_t c = { 0 };
	//init();
	// Generate clock signal on OC1A (PB5 aka D11 on Mega)
	DDRB = _BV(7) | _BV(6) | _BV(5);
	TCCR1A = _BV(COM1A0) | _BV(WGM11) | _BV(WGM10);
	TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);	// fclk/1
	OCR1A = 0xF3;	// should generate 32786.8859 Hz
	
	// set up UART0 for transmitting
	
	UCSR0A |= _BV(U2X0);
	UCSR0B |= _BV(TXEN0);
	UCSR0C |= _BV(UCSZ01) | _BV(UCSZ02);
	UBRR0 = 7;	// 250 kbps
	
	// set up SPI for emulating MS5541C's protocol
	SPI_DDR &= ~(_BV(SPI_MOSI)|_BV(SPI_MISO)|_BV(SPI_SS)|_BV(SPI_SCK));
	// Define the following pins as output
	SPI_DDR |= (_BV(SPI_MOSI) | _BV(SPI_SS) | _BV(SPI_SCK));
	SPI_PORT |= _BV(SPI_SS);
	SPCR = _BV(SPE) | _BV(MSTR) | _BV(SPR1);	// enable SPI, set as master, set prescaler to f(osc)/64
	SPSR = _BV(SPI2X);							// Double SCK to f(osc)/32 (0.5 MHz)
	
	// SCK - D9, PH6
	// DOUT - D8, PH5 (MISO)
	// DIN - D7, PH4  (MOSI)
	//DDRH = _BV(PH6) | _BV(PH4);
	
	//PORTB &= ~_BV(PB6);
	
	//send_bits(RESET, RESET_BITS);
	send_bytes(RESET);
	send_bytes(0);
	get_coeffs(&c);
	print("C1 = SENST1 = %u\r\nC2 = OFFT1 = %u\r\nC3 = TCS = %u\r\nC4 = TCO = %u\r\nC5 = Tref = %u\r\nC6 = TEMPSENS = %u\r\n", c.SENST1, c.OFFT1, c.TCS, c.TCO, c.Tref, c.TEMPSENS);
	for (;;)
	{
		dT = get_temp_diff(&c);
		v = get_temperature(dT, &c);
		print("Temperature reading = %u.%u degrees\r\n", v/10, v%10);
		v = get_pressure(dT, &c);
		print("Pressure reading = %u mbar\r\n", v);
		_delay_ms(1000);
	}
    while(1)
    {
        //print("bleargh!\r\n");
		//_delay_ms(1000);

    }
}